package threads;

import java.util.concurrent.locks;

public class Buffer2<T>
{
	T[] storage;
	int head, tail, numEl;

	Lock lock;
	Condition empty, full;

	public Buffer2(int size)
	{
		storage = (T[]) new Object[size];
		head = tail = numEl = 0;
		lock = new ReentrantLock();
		full = lock.newCondition();
		empty = lock.newCondition();
	}

	public Buffer2()
	{
		this(10);
	}

	public boolean isEmpty()
	{
		return size() == 0;
		// non e' meno efficiente di numEl == 0 perche' Java usa pesantemente
		// l'ottimizzazione di inlining
	}

	public boolean isFull()
	{
		return size() == storage.length;
	}

	public int size()
	{
		return numEl;
	}

	public T get()
	{
		lock.lock();
		while(isEmpty())
		{
			// attesa qui
			try
			{
				empty.await();
			}
			catch(InterruptedException ie)
			{
				ie.printStackTrace();
			}
		}

		T retval = storage[head];
		head = (head+1)%storage.length;
		numEl--;

		full.signal();

		lock.unlock();
		return retval;
	}

	public void put(T elem)
	{
		lock.lock();

		while(isFull())
		{
			//qui attesa
			try
			{
				full.await();
			}
			catch(InterruptedException ie)
			{
				ie.printStackTrace();
			}
		}

		storage[tail] = elem;
		tail = (tail+1)%storage.length;
		numEl++;
		
		empty.signal();

		lock.unlock();
	}
}
